import asyncio
import re
from urllib.request import url2pathname
from pathlib import Path
from typing import Any, Callable, Optional
from collections.abc import Coroutine

import filetype
from nonebot import logger
from nonebot.adapters import Bot
from nonebot.adapters.discord import Bot as dc_Bot
from nonebot.adapters.discord.api import Embed, EmbedAuthor, File
from nonebot.adapters.discord.exception import NetworkError
from nonebot.adapters.onebot.v11 import (
    Bot as qq_Bot,
    GroupMessageEvent,
    GroupRecallNoticeEvent,
    Message,
    MessageSegment,
)
from nonebot.adapters.onebot.v11.event import Reply
from nonebot_plugin_orm import get_session
from sqlalchemy import select

from .config import LinkWithWebhook
from .qq_emoji_dict import qq_emoji_dict
from .model import MsgID
from .utils import get_file_bytes, skil_to_ogg


async def get_qq_member_name(bot: qq_Bot, group_id: int, user_id: int) -> str:
    return (
        await bot.get_group_member_info(
            group_id=group_id, user_id=user_id, no_cache=True
        )
    )["nickname"]


async def build_dc_file(bot: Bot, file: str, url: str) -> File:
    """获取文件，用于发送到 Discord"""
    img_bytes = await get_file_bytes(bot, url)
    if re.search(r"\.[a-z]+$", file):
        filename = file
    else:
        match = filetype.match(img_bytes)
        filename = file + ("." + match.extension) if match else ""
    return File(content=img_bytes, filename=filename)


async def create_qq_to_dc(
    bot: qq_Bot,
    event: GroupMessageEvent,
    dc_bot: dc_Bot,
    channel_links: list[LinkWithWebhook],
):
    """QQ 消息转发到 discord"""
    logger.debug("create qq to dc: start")
    link = next(link for link in channel_links if link.qq_group_id == event.group_id)
    builder = MessageBuilder()

    seg_msg = event.get_message()
    text, files, embeds = await builder.build(seg_msg, bot, event)

    if reply := event.reply:
        embeds = [await builder.handle_reply(reply, bot, link), *embeds]

    username = (
        f"{event.sender.card or event.sender.nickname} [QQ:{event.sender.user_id}]"
    )
    avatar = f"https://q.qlogo.cn/g?b=qq&nk={event.sender.user_id}&s=100"

    for try_times in range(3):
        try:
            send = await dc_bot.execute_webhook(
                webhook_id=link.webhook_id,
                token=link.webhook_token,
                content=text,
                files=files,
                embeds=embeds,
                username=username,
                avatar_url=avatar,
                wait=True,
            )
            break
        except (NameError, NetworkError) as e:
            logger.warning(f"create qq to dc error: {e}, retry {try_times + 1}")
            if try_times >= 2:
                continue
            await asyncio.sleep(5)
    else:
        logger.error("create qq to dc: failed")
        return

    async with get_session() as session:
        session.add(MsgID(dcid=send.id, qqid=event.message_id))
        await session.commit()

    logger.debug("create qq to dc: done")


async def delete_qq_to_dc(
    event: GroupRecallNoticeEvent,
    dc_bot: dc_Bot,
    channel_links: list[LinkWithWebhook],
    just_delete: list,
):
    logger.debug("delete qq to dc: start")
    if (id := event.message_id) in just_delete:
        just_delete.remove(id)
        return
    channel_id = next(
        link.dc_channel_id
        for link in channel_links
        if link.qq_group_id == event.group_id
    )
    for try_times in range(3):
        try:
            async with get_session() as session:
                if msgids := await session.scalars(
                    select(MsgID).filter(MsgID.qqid == event.message_id)
                ):
                    for msgid in msgids:
                        await dc_bot.delete_message(
                            message_id=msgid.dcid, channel_id=channel_id
                        )
                        just_delete.append(msgid.dcid)
                        await session.delete(msgid)
                    await session.commit()
            logger.debug("delete qq to dc: done")
            break
        except (UnboundLocalError, TypeError, NameError) as e:
            logger.warning(f"delete qq to dc error: {e}, retry {try_times + 1}")
            if try_times >= 2:
                continue
            await asyncio.sleep(5)
    else:
        logger.error("delete qq to dc: failed")


class MsgResult:
    """消息处理结果容器"""

    embed: Optional[Embed]
    ensure_text: bool
    file: Optional[File]
    text: Optional[str]

    __slots__ = ("embed", "ensure_text", "file", "text")

    def __init__(self, *, embed=None, ensure=False, file=None, text=None):
        self.embed = embed
        self.ensure_text = ensure
        self.file = file
        self.text = text


class MessageBuilder:
    _mapping: dict[
        str,
        Callable[
            [MessageSegment, qq_Bot, GroupMessageEvent],
            Coroutine[Any, Any, MsgResult],
        ],
    ]

    def __init__(self):
        self._mapping = {
            "text": self.text,
            "face": self.face,
            "mface": self.mface,
            "marketface": self.marketface,
            "at": self.at,
            "image": self.image,
            "record": self.record,
            "video": self.video,
            "file": self.file,
            "share": self.share,
            "contact": self.contact,
            "location": self.location,
            "music": self.music,
            "forward": self.forward,
            "xml": self.xml,
            "json": self.json,
        }

    async def build(
        self, seg_msg: Message, bot: qq_Bot, event: GroupMessageEvent
    ) -> tuple[str, list[File], list[Embed]]:
        result: list[Coroutine[Any, Any, MsgResult]] = []
        text: str = ""
        files: list[File] = []
        embeds: list[Embed] = []

        result.extend(self.convert(seg, bot, event) for seg in seg_msg)

        send_msg = await asyncio.gather(*result)

        for res in send_msg:
            text += (
                res.text
                if res.text and (res.ensure_text or seg_msg.extract_plain_text())
                else ""
            )
            files.append(res.file) if res.file else ...
            embeds.append(res.embed) if res.embed else ...

        if not (text or files or embeds):
            text = "[未知消息]"

        return text, files, embeds

    def convert(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> Coroutine[Any, Any, MsgResult]:
        seg_type = seg.type
        if seg_type in self._mapping:
            res = self._mapping[seg_type](seg, bot, event)
        else:
            res = self.other(seg, bot, event)

        return res

    def extract_plain_text(self, msg_res: list[MsgResult]) -> str:
        return "".join([res.text for res in msg_res if res.text])

    async def handle_reply(
        self,
        reply: Reply,
        bot: qq_Bot,
        link: LinkWithWebhook,
    ) -> Embed:
        author = ""
        sender = reply.sender
        timestamp = f"<t:{reply.time}:R>"
        plaintext_msg = reply.message.extract_plain_text()
        guild_id, channel_id = link.dc_guild_id, link.dc_channel_id

        if (
            re.match(r"^.*?\(@.*?\):\n\n", plaintext_msg)
            and str(sender.user_id) == bot.self_id
        ):
            name = plaintext_msg.split("\n")[0][:-1]
            author = EmbedAuthor(name=name)
            plaintext_msg = "\n".join(plaintext_msg.split("\n")[2:])
        else:
            author = EmbedAuthor(
                name=(sender.card or sender.nickname or "") + f"[QQ:{sender.user_id}]",
                icon_url=f"https://q.qlogo.cn/g?b=qq&nk={sender.user_id}&s=100",
            )

        async with get_session() as session:
            reference_id = await session.scalar(
                select(MsgID.dcid).filter(MsgID.qqid == reply.message_id).limit(1)
            )
        if reference_id:
            description = (
                f"{plaintext_msg}\n\n"
                + timestamp
                + f"[[ ↑ ]](https://discord.com/channels/{guild_id}/{channel_id}/{reference_id})"
            )
        else:
            description = f"{plaintext_msg}\n\n{timestamp}[ ? ]"

        return Embed(
            author=author,
            description=description,
        )

    async def text(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        content = (
            seg.data["text"]
            .replace("@everyone", "@.everyone")
            .replace("@here", "@.here")
        )
        return MsgResult(text=content)

    async def at(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        data = seg.data
        qq = data.get("user_id") or data["qq"]

        if qq in ["0", "all"]:
            return MsgResult(text="@everyone")

        name = data.get("name")
        if not name:
            name = await get_qq_member_name(bot, event.group_id, qq)

        return MsgResult(text=f"[{name}](mailto:{qq}@qq.com)[QQ:{qq}] ")

    async def face(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        face_id = str(seg.data["id"])
        emoji = qq_emoji_dict.get(face_id, f"QQemojiID:{face_id}")
        return MsgResult(text=f"[{emoji}]")

    async def mface(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        data = seg.data

        if data.get("summary") and data.get("url"):
            return MsgResult(
                text=data["summary"],
                file=await build_dc_file(bot, f"{data['summary']}.gif", data["url"]),
            )

        return MsgResult(
            text="[动画表情]",
            file=await build_dc_file(
                bot,
                f"{data['id']}.gif",
                "https://gxh.vip.qq.com/club/item/parcel/item/"
                + f"{data['id'][:2]}/{data['id']}/raw300.gif",
            ),
        )

    async def marketface(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        return MsgResult(
            text=seg.data["summary"],
            file=await build_dc_file(
                bot,
                f"{seg.data['summary']}.gif",
                "https://gxh.vip.qq.com/club/item/parcel/item/"
                + f"{seg.data['face_id'][:2]}/{seg.data['face_id']}/raw300.gif",
            ),
        )

    async def image(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        return MsgResult(
            text="[图片]",
            file=await build_dc_file(bot, seg.data["file"][-40:], seg.data["url"]),
        )

    async def record(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        if path := seg.data.get("path", ""):
            while True:
                if not Path(path).exists():
                    await asyncio.sleep(0.1)
                    continue
                record_bytes = Path(path).read_bytes()
                break
        elif url := seg.data.get("url", ""):
            record_bytes = await get_file_bytes(bot, url)
        else:
            path = url2pathname(seg.data["file"].removeprefix("file:"))
            record_bytes = Path(path).read_bytes()

        return MsgResult(
            text="[语音]",
            file=File(
                content=skil_to_ogg(record_bytes),
                filename="voice-message.ogg",
            ),
        )

    async def video(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        filename = seg.data.get("file_name", "") or seg.data.get("file", "")
        url = seg.data["url"]

        if path := seg.data.get("path", ""):
            content = Path(path).read_bytes()
        elif re.match(r"[A-Z]:\\\\", url):
            content = Path(url).read_bytes()
        elif "http" in url:
            content = await get_file_bytes(bot, url)
        else:
            return MsgResult(text=f"[{filename}]")

        return MsgResult(
            text=f"[{filename}]",
            file=File(filename=filename, content=content),
        )

    async def file(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        filename = (
            seg.data.get("file_name", "")
            or seg.data.get("name", "")
            or seg.data.get("file", "")
        )

        if "http" in seg.data.get("url", ""):
            content = await get_file_bytes(bot, seg.data["url"])
        elif file_id := seg.data.get("file_id", ""):
            file_info = await bot.call_api("get_file", file_id=file_id)
            content = Path(file_info["file"]).read_bytes()
        else:
            return MsgResult(text=f"[{filename}]")

        return MsgResult(
            text=f"[{filename}]", file=File(content=content, filename=filename)
        )

    async def share(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        data = seg.data
        embed = Embed(
            title=data["title"],
            url=data["url"],
            description=data["content"],
            image=data["image"],
        )
        return MsgResult(embed=embed)

    async def contact(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        contact_type = "好友" if seg.data["type"] == "qq" else "群"
        return MsgResult(ensure=True, text=f"<推荐{contact_type}：{seg.data['id']}>")

    async def location(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        data = seg.data
        embed = Embed(
            title=data["title"],
            description=f"{data['content']}\nlat:{data['lat']}, lon: {data['lon']}",
        )
        return MsgResult(embed=embed)

    async def music(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        return MsgResult(ensure=True, text="[音乐分享]")

    async def forward(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        return MsgResult(ensure=True, text="[合并转发]")

    async def xml(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        return (
            MsgResult(ensure=True, text=f"[xml 消息]({seg.data})")
            if len(seg) == 1
            else MsgResult()
        )

    async def json(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        return (
            MsgResult(ensure=True, text=f"[json 消息]({seg.data})")
            if len(seg) == 1
            else MsgResult()
        )

    async def rps(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        res = {"0": "", "1": "石头", "2": "剪刀", "3": "布"}[
            seg.data.get("result", "0")
        ]
        return MsgResult(ensure=True, text=f"[猜拳{'：' if res else ''}{res}]")

    async def dice(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        res = seg.data.get("result", "")
        return MsgResult(ensure=True, text=f"[掷骰子{'：' if res else ''}{res}]")

    async def other(
        self, seg: MessageSegment, bot: qq_Bot, event: GroupMessageEvent
    ) -> MsgResult:
        return MsgResult(
            ensure=True, text=f"[不支持的类型](type: {seg.type}, data: {seg.data})"
        )



"""init db

迁移 ID: ca5c07252665
父迁移:
创建时间: 2024-05-24 16:05:18.822389

"""

from __future__ import annotations

from collections.abc import Sequence

from alembic import op
import sqlalchemy as sa


revision: str = "ca5c07252665"
down_revision: str | Sequence[str] | None = None
branch_labels: str | Sequence[str] | None = ("nonebot_plugin_dcqq_relay",)
depends_on: str | Sequence[str] | None = None


def upgrade(name: str = "") -> None:
    if name:
        return
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "nonebot_plugin_dcqq_relay_msgid",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("dcid", sa.Integer(), nullable=False),
        sa.Column("qqid", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_nonebot_plugin_dcqq_relay_msgid")),
        info={"bind_key": "nonebot_plugin_dcqq_relay"},
    )
    # ### end Alembic commands ###


def downgrade(name: str = "") -> None:
    if name:
        return
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("nonebot_plugin_dcqq_relay_msgid")
    # ### end Alembic commands ###



"""change dcid type

迁移 ID: 5fb7a7432778
父迁移: ca5c07252665
创建时间: 2025-07-06 16:40:46.261312

"""

from __future__ import annotations

from collections.abc import Sequence

from alembic import op
import sqlalchemy as sa

revision: str = "5fb7a7432778"
down_revision: str | Sequence[str] | None = "ca5c07252665"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade(name: str = "") -> None:
    if name:
        return
    # ### commands auto generated by Alembic - please adjust! ###
    bind = op.get_bind()
    if bind.engine.name == "sqlite":
        with op.batch_alter_table("nonebot_plugin_dcqq_relay_msgid") as op_batch:
            op_batch.alter_column("dcid", nullable=False, type_=sa.BigInteger())
    else:
        op.alter_column(
            "nonebot_plugin_dcqq_relay_msgid",
            "dcid",
            nullable=False,
            type_=sa.BigInteger(),
        )
    # ### end Alembic commands ###


def downgrade(name: str = "") -> None:
    if name:
        return
    # ### commands auto generated by Alembic - please adjust! ###
    bind = op.get_bind()
    if bind.engine.name == "sqlite":
        with op.batch_alter_table("nonebot_plugin_dcqq_relay_msgid") as op_batch:
            op_batch.alter_column(
                "dcid",
                nullable=False,
                type_=sa.Integer(),
                existing_type=sa.BigInteger(),
            )
    else:
        op.alter_column(
            "nonebot_plugin_dcqq_relay_msgid",
            "dcid",
            nullable=False,
            type_=sa.Integer(),
            existing_type=sa.BigInteger(),
        )
    # ### end Alembic commands ###



import re
import ssl
from typing import Optional, Union

import pysilk
from pydub import AudioSegment
from io import BytesIO

from nonebot import logger
from nonebot.adapters import Bot
from nonebot.internal.driver import Request
from nonebot.adapters.discord import (
    Bot as dc_Bot,
    GuildMessageCreateEvent,
    GuildMessageDeleteEvent,
)
from nonebot.adapters.discord.api import UNSET, Missing
from nonebot.adapters.discord.exception import ActionFailed
from nonebot.adapters.onebot.v11 import (
    GroupMessageEvent,
    GroupRecallNoticeEvent,
)
from nonebot.compat import model_dump

from .config import LinkWithoutWebhook, LinkWithWebhook, plugin_config

channel_links: list[LinkWithoutWebhook] = plugin_config.dcqq_relay_channel_links
discord_proxy = plugin_config.discord_proxy


async def check_messages(
    event: Union[
        GroupMessageEvent,
        GuildMessageCreateEvent,
        GroupRecallNoticeEvent,
        GuildMessageDeleteEvent,
    ],
) -> bool:
    """检查消息"""
    logger.debug("into check_messages()")
    if isinstance(event, GroupMessageEvent):
        return any(event.group_id == link.qq_group_id for link in channel_links)
    elif isinstance(event, GuildMessageCreateEvent):
        if not (
            re.match(r".*?\[QQ:\d*?\]$", event.author.username)
            and event.author.bot is True
        ):
            return any(
                event.guild_id == link.dc_guild_id
                and event.channel_id == link.dc_channel_id
                for link in channel_links
            )
        else:
            return False
    elif isinstance(event, GroupRecallNoticeEvent):
        return any(event.group_id == link.qq_group_id for link in channel_links)
    elif isinstance(event, GuildMessageDeleteEvent):
        return any(
            event.guild_id == link.dc_guild_id
            and event.channel_id == link.dc_channel_id
            for link in channel_links
        )


async def check_to_me(
    event: Union[
        GroupMessageEvent,
        GuildMessageCreateEvent,
        GroupRecallNoticeEvent,
        GuildMessageDeleteEvent,
    ],
) -> bool:
    if isinstance(event, (GroupMessageEvent, GuildMessageCreateEvent)):
        return event.to_me
    return True


async def get_dc_member_name(
    bot: dc_Bot, guild_id: Missing[int], user_id: int
) -> tuple[str, str]:
    try:
        if guild_id is not UNSET:
            member = await bot.get_guild_member(guild_id=guild_id, user_id=user_id)
            if (nick := member.nick) and nick is not UNSET:
                return nick, member.user.username if member.user is not UNSET else ""
            elif member.user is not UNSET and (global_name := member.user.global_name):
                return global_name, member.user.username
            else:
                return "", str(user_id)
        else:
            user = await bot.get_user(user_id=user_id)
            return user.global_name or "", user.username
    except ActionFailed as e:
        if e.message == "Unknown User":
            return "(error:未知用户)", str(user_id)
        else:
            raise e


async def get_file_bytes(bot: Bot, url: str, proxy: Optional[str] = None) -> bytes:
    try:
        resp = await bot.adapter.request(Request("GET", url, proxy=proxy))
        if isinstance(resp.content, bytes):
            return resp.content
        else:
            raise TypeError("Response content is not bytes")
    except ssl.SSLError as e:
        if url.startswith("http://"):
            raise e
        url = url.replace("https://", "http://")
        return await get_file_bytes(bot, url, proxy)


async def get_webhook(
    bot: dc_Bot, link: LinkWithoutWebhook
) -> Union[LinkWithWebhook, int]:
    if link.webhook_id and link.webhook_token:
        return LinkWithWebhook(**model_dump(link))
    try:
        channel_webhooks = await bot.get_channel_webhooks(channel_id=link.dc_channel_id)
        bot_webhook = next(
            (
                webhook
                for webhook in channel_webhooks
                if webhook.application_id == int(bot.self_id)
            ),
            None,
        )
        if bot_webhook and bot_webhook.token:
            return build_link(link, bot_webhook.id, bot_webhook.token)
    except Exception as e:
        logger.error(
            f"get webhook error, Discord channel id: {link.dc_channel_id}, error: {e}"
        )
    try:
        create_webhook = await bot.create_webhook(
            channel_id=link.dc_channel_id, name=str(link.dc_channel_id)
        )
        if create_webhook.token:
            return build_link(link, create_webhook.id, create_webhook.token)
    except Exception as e:
        logger.error(
            f"create webhook error, Discord channel id: {link.dc_channel_id}, "
            + f"error: {e}"
        )
    logger.error(
        f"failed to get or create webhook, Discord channel id: {link.dc_channel_id}"
    )
    return link.dc_channel_id


def build_link(
    link: LinkWithoutWebhook, webhook_id: int, webhook_token: str
) -> LinkWithWebhook:
    return LinkWithWebhook(
        webhook_id=webhook_id,
        webhook_token=webhook_token,
        **model_dump(link, exclude={"webhook_id", "webhook_token"}),
    )


def pydub_transform(origin_bytes: bytes, input_type: str, output_type: str) -> bytes:
    output_buffer = BytesIO()  # 创建内存文件对象

    audio = AudioSegment.from_file(BytesIO(origin_bytes), format=input_type)
    audio.export(output_buffer, format=output_type, bitrate="128k")

    output_buffer.seek(0)  # 重置指针
    return output_buffer.read()


def skil_to_ogg(skil_bytes: bytes) -> bytes:
    output_buffer = BytesIO()

    pcm_bytes = pysilk.decode(skil_bytes, True, sample_rate=44100)
    audio = AudioSegment.from_file(BytesIO(pcm_bytes), format="wav")
    audio.export(output_buffer, format="ogg")

    output_buffer.seek(0)
    return output_buffer.read()


async def get_dc_member_avatar(bot: dc_Bot, guild_id: int, user_id: int) -> str:
    member = await bot.get_guild_member(guild_id=guild_id, user_id=user_id)
    if member.avatar is not UNSET and (avatar := member.avatar):
        return (
            f"https://cdn.discordapp.com/guilds/{guild_id}/users/{user_id}/avatars/{avatar}."
            + ("gif" if re.match(r"^a_.*", avatar) else "webp")
        )
    elif (user := member.user) and user is not UNSET and user.avatar:
        return f"https://cdn.discordapp.com/avatars/{user_id}/{user.avatar}." + (
            "gif" if re.match(r"^a_.*", user.avatar) else "webp"
        )
    else:
        return ""



import asyncio
from typing import Any, Callable, Optional
from collections.abc import Coroutine
from pathlib import Path

from nonebot import logger
from nonebot.adapters.discord import (
    Bot as dc_Bot,
    GuildMessageCreateEvent,
    GuildMessageDeleteEvent,
)
from nonebot.adapters.discord.message import (
    MessageSegment as dc_MS,
    Message as dc_M,
)
from nonebot.adapters.discord.api import UNSET, Attachment, StickerItem, MessageGet
from nonebot.adapters.onebot.v11 import (
    Bot as qq_Bot,
    Message as qq_M,
    MessageSegment as qq_MS,
)
from nonebot.adapters.onebot.utils import f2s
from nonebot_plugin_orm import get_session
from nonebot_plugin_localstore import get_plugin_cache_dir
from sqlalchemy import select

from .config import LinkWithWebhook, plugin_config
from .model import MsgID
from .utils import get_dc_member_name, get_file_bytes, pydub_transform

discord_proxy = plugin_config.discord_proxy
cache_dir = get_plugin_cache_dir()


async def get_dc_channel_name(bot: dc_Bot, guild_id: int, channel_id: int) -> str:
    channels = await bot.get_guild_channels(guild_id=guild_id)
    channel = next(channel for channel in channels if channel.id == channel_id)
    return (
        channel.name
        if channel.name is not UNSET and channel.name is not None
        else "(error:无名频道)"
    )


async def get_dc_role_name(bot: dc_Bot, guild_id: int, role_id: int) -> str:
    roles = await bot.get_guild_roles(guild_id=guild_id)
    role = next(role for role in roles if role.id == role_id)
    return role.name


async def upload_group_file(bot: qq_Bot, group_id: int, file: qq_M):
    await bot.call_api(
        "upload_group_file",
        group_id=group_id,
        file=file[0].data["file"],
        name=file[0].data["name"],
        folder="",
    )


async def prepare_file(bot: qq_Bot, files: list[qq_M]) -> tuple[list[qq_M], bool]:
    need_upload: bool = False
    version_info = await bot.get_version_info()

    if version_info["app_name"] == "NapCat.Onebot":
        for file in files:
            file[0].data["file"] = f2s(file[0].data["file"])
        return files, need_upload

    if version_info["app_name"] == "Lagrange.OneBot":
        need_upload = True
    for file in files:
        file[0].data["file"] = (
            save_file(file[0].data["file"], file[0].data["name"])
        ).as_posix()
    return files, need_upload


def save_file(file: bytes, file_name: str) -> Path:
    file_path = cache_dir / file_name
    file_path.write_bytes(file)
    return file_path


async def ensure_message(bot: dc_Bot, event: GuildMessageCreateEvent) -> dc_M:
    attrs = ("attachments", "content", "embeds", "components", "sticker_items")
    if not any(getattr(event, attr) for attr in attrs):
        message_get = await bot.get_channel_message(
            channel_id=event.channel_id, message_id=event.message_id
        )
        for attr in attrs:
            setattr(event, attr, getattr(message_get, attr))
    return event.get_message()


def split_messages(messages: qq_M) -> tuple[list[qq_M], list[qq_M]]:
    combinable, videos, files = qq_M(), [], []
    for seg in messages:
        if seg.type in ["text", "image", "reply", "record"]:
            combinable.append(seg)
        elif seg.type == "video":
            videos.append(qq_M(seg))
        elif seg.type == "file":
            files.append(qq_M(seg))
    return [combinable, *videos], files


async def gather_send(
    bot: qq_Bot,
    qq_group_id: int,
    msg_to_send: list[qq_M],
    files: list[qq_M],
) -> list[dict[str, Any]]:
    tasks = []
    files, need_upload = await prepare_file(bot, files)

    if need_upload:
        tasks.extend(upload_group_file(bot, qq_group_id, file) for file in files)
    else:
        msg_to_send += files
    tasks = [
        bot.send_group_msg(group_id=qq_group_id, message=message)
        for message in msg_to_send
    ] + tasks

    sends = await asyncio.gather(*tasks)
    return [send for send in sends if send is not None]


async def create_dc_to_qq(
    bot: dc_Bot,
    event: GuildMessageCreateEvent,
    qq_bot: qq_Bot,
    channel_links: list[LinkWithWebhook],
):
    """discord 消息转发到 QQ"""
    logger.debug("create dc to qq: start")
    link = next(
        link for link in channel_links if link.dc_channel_id == event.channel_id
    )
    seg_msg = await ensure_message(bot, event)

    messages = await MessageBuilder().build(seg_msg, bot, event)
    msg_to_send, files = split_messages(messages)

    for try_times in range(3):
        try:
            sends = await gather_send(
                qq_bot,
                link.qq_group_id,
                msg_to_send,
                files,
            )
            break
        except NameError as e:
            logger.warning(f"create dc to qq error: {e}, retry {try_times + 1}")
            if try_times == 2:
                continue
            await asyncio.sleep(5)
    else:
        logger.error("create dc to qq: failed")
        return

    async with get_session() as session:
        session.add_all(MsgID(dcid=event.id, qqid=send["message_id"]) for send in sends)
        await session.commit()

    logger.debug("create dc to qq done")


async def delete_dc_to_qq(
    event: GuildMessageDeleteEvent,
    qq_bot: qq_Bot,
    just_delete: list,
):
    logger.debug("delete dc to qq: start")
    if (id := event.id) in just_delete:
        just_delete.remove(id)
        return
    for try_times in range(3):
        try:
            async with get_session() as session:
                if msgids := await session.scalars(
                    select(MsgID).filter(MsgID.dcid == event.id)
                ):
                    for msgid in msgids:
                        await qq_bot.delete_msg(message_id=msgid.qqid)
                        just_delete.append(msgid.qqid)
                        await session.delete(msgid)
                    await session.commit()
            logger.debug("delete dc to qq: done")
            break
        except (UnboundLocalError, TypeError, NameError) as e:
            logger.warning(f"delete dc to qq error: {e}, retry {try_times + 1}")
            if try_times == 2:
                continue
            await asyncio.sleep(5)
    else:
        logger.warning("delete dc to qq: failed")


class MessageBuilder:
    _mapping: dict[
        str,
        Callable[
            [dc_MS, dc_Bot, GuildMessageCreateEvent],
            Coroutine[Any, Any, Optional[qq_MS]],
        ],
    ]

    def __init__(self):
        self._mapping = {
            "attachment": self.attachment,
            "sticker": self.sticker,
            "embed": self.embed,
            "component": self.component,
            "custom_emoji": self.custom_emoji,
            "mention_user": self.mention_user,
            "mention_role": self.mention_role,
            "mention_channel": self.mention_channel,
            "mention_everyone": self.mention_everyone,
            "text": self.text,
            "timestamp": self.timestamp,
        }

    async def build(
        self, seg_msg: dc_M, bot: dc_Bot, event: GuildMessageCreateEvent
    ) -> qq_M:
        result: list[Coroutine[Any, Any, Optional[qq_MS]]] = [self.build_sender(event)]

        if event.content or event.embeds or event.components:
            result.extend(self.convert(seg, bot, event) for seg in seg_msg)

        result.extend(
            self.handle_attachment(attachment, bot, event)
            for attachment in event.attachments
        )

        if event.sticker_items:
            result.extend(
                self.handle_sticker(sticker) for sticker in event.sticker_items
            )

        if referenced_message := event.referenced_message:
            result.append(self.handle_referenced_message(referenced_message))

        send_msg = await asyncio.gather(*result)

        return qq_M([seg for seg in send_msg if seg is not None])

    def convert(
        self, seg: dc_MS, bot: dc_Bot, event: GuildMessageCreateEvent
    ) -> Coroutine[Any, Any, Optional[qq_MS]]:
        seg_type = seg.type
        if seg_type in self._mapping:
            res = self._mapping[seg_type](seg, bot, event)
        else:
            res = self.other(seg, bot, event)

        return res

    async def handle_attachment(
        self, attachment: Attachment, bot: dc_Bot, event: GuildMessageCreateEvent
    ) -> qq_MS:
        filename = attachment.filename
        content_type = (
            attachment.content_type if attachment.content_type is not UNSET else ""
        )
        filetype = filename.split(".")[-1]
        if "image" in content_type:
            return qq_MS.image(
                await get_file_bytes(bot, attachment.url, discord_proxy), type_=filetype
            )
        if "video" in content_type:
            return qq_MS.video(await get_file_bytes(bot, attachment.url, discord_proxy))
        if "audio" in content_type and hasattr(attachment, "duration_secs"):
            return qq_MS.record(
                pydub_transform(
                    await get_file_bytes(bot, attachment.url, discord_proxy),
                    filetype,
                    "mp3",
                )
            )
        return qq_MS(
            "file",
            {
                "file": await get_file_bytes(bot, attachment.url, discord_proxy),
                "name": filename,
            },
        )

    async def handle_sticker(self, sticker: StickerItem) -> Optional[qq_MS]:
        return qq_MS.text(f"[{sticker.name}]")
        # WIP

    async def handle_referenced_message(
        self, referenced: MessageGet
    ) -> Optional[qq_MS]:
        async with get_session() as session:
            if reply_id := await session.scalar(
                select(MsgID.qqid).filter(MsgID.dcid == referenced.id).limit(1)
            ):
                return qq_MS.reply(reply_id)

    async def build_sender(self, event: GuildMessageCreateEvent) -> qq_MS:
        return qq_MS.text(
            (
                event.member.nick
                if event.member is not UNSET
                and event.member.nick is not UNSET
                and event.member.nick
                else event.author.global_name or ""
            )
            + f"(@{event.author.username}):\n\n"
        )

    async def attachment(
        self, seg: dc_MS, bot: dc_Bot, event: GuildMessageCreateEvent
    ) -> Optional[qq_MS]:
        filename = seg.data["attachment"].filename
        if hasattr(seg.data["attachment"], "duration_secs"):
            return
        else:
            return qq_MS.text(f"[{filename}]" if len(filename) < 10 else "[文件]")

    async def sticker(
        self, seg: dc_MS, bot: dc_Bot, event: GuildMessageCreateEvent
    ) -> qq_MS:
        return qq_MS.image(
            await get_file_bytes(
                bot,
                f"https://cdn.discordapp.com/stickers/{seg.data['id']}.gif",
                discord_proxy,
            )
        )

    async def embed(
        self, seg: dc_MS, bot: dc_Bot, event: GuildMessageCreateEvent
    ) -> qq_MS:
        return qq_MS.text(f"\n<Embed:{seg.data['embed'].title}>\n")

    async def component(
        self, seg: dc_MS, bot: dc_Bot, event: GuildMessageCreateEvent
    ) -> qq_MS:
        return qq_MS.text(f"\n{seg}\n")

    async def custom_emoji(
        self, seg: dc_MS, bot: dc_Bot, event: GuildMessageCreateEvent
    ) -> qq_MS:
        return qq_MS.image(
            await get_file_bytes(
                bot,
                "https://cdn.discordapp.com/emojis/"
                + seg.data["id"]
                + "."
                + ("gif" if seg.data["animated"] is True else "webp"),
                discord_proxy,
            )
        )

    async def mention_user(
        self, seg: dc_MS, bot: dc_Bot, event: GuildMessageCreateEvent
    ) -> qq_MS:
        nick, username = await get_dc_member_name(
            bot, event.guild_id, int(seg.data["user_id"])
        )
        return qq_MS.text("@" + nick + f"({username})")

    async def mention_role(
        self, seg: dc_MS, bot: dc_Bot, event: GuildMessageCreateEvent
    ) -> qq_MS:
        return qq_MS.text(
            "@" + await get_dc_role_name(bot, event.guild_id, int(seg.data["role_id"]))
        )

    async def mention_channel(
        self, seg: dc_MS, bot: dc_Bot, event: GuildMessageCreateEvent
    ) -> qq_MS:
        return qq_MS.text(
            "#"
            + await get_dc_channel_name(
                bot, event.guild_id, int(seg.data["channel_id"])
            )
        )

    async def mention_everyone(
        self, seg: dc_MS, bot: dc_Bot, event: GuildMessageCreateEvent
    ) -> qq_MS:
        return qq_MS.at("all")

    async def text(
        self, seg: dc_MS, bot: dc_Bot, event: GuildMessageCreateEvent
    ) -> qq_MS:
        return qq_MS.text(seg.data["text"])

    async def timestamp(
        self, seg: dc_MS, bot: dc_Bot, event: GuildMessageCreateEvent
    ) -> qq_MS:
        return qq_MS.text(str(seg))

    async def other(
        self, seg: dc_MS, bot: dc_Bot, event: GuildMessageCreateEvent
    ) -> qq_MS:
        return qq_MS.text(f"\n{seg}\n")



import asyncio
from typing import Union

from nonebot import get_driver, logger, on, require
from nonebot.adapters import Event
from nonebot.adapters.discord import (
    Bot as dc_Bot,
    GuildMessageCreateEvent,
    GuildMessageDeleteEvent,
)
from nonebot.adapters.onebot.v11 import (
    Bot as qq_Bot,
    GroupMessageEvent,
    GroupRecallNoticeEvent,
)
from nonebot.plugin import PluginMetadata
from nonebot.rule import Rule, StartswithRule
from nonebot.typing import T_State

require("nonebot_plugin_orm")
require("nonebot_plugin_localstore")

from .config import Config, LinkWithoutWebhook, LinkWithWebhook, plugin_config
from .dc_to_qq import create_dc_to_qq, delete_dc_to_qq
from .qq_to_dc import create_qq_to_dc, delete_qq_to_dc
from .utils import check_messages, check_to_me, get_webhook

__plugin_meta__ = PluginMetadata(
    name="QQ群-Discord 互通",
    description="在QQ群与 Discord 之间同步消息的 nonebot2 插件",
    usage="",
    type="application",
    homepage="https://github.com/Autuamn/nonebot-plugin-dcqq-relay",
    config=Config,
    supported_adapters={"~onebot.v11", "~discord"},
)


driver = get_driver()
without_webhook_links: list[LinkWithoutWebhook] = plugin_config.dcqq_relay_channel_links
with_webhook_links: list[LinkWithWebhook] = []
discord_proxy = plugin_config.discord_proxy
unmatch_beginning = plugin_config.dcqq_relay_unmatch_beginning

just_delete = []


class NotStartswithRule(StartswithRule):
    async def __call__(self, event: Event, state: T_State) -> bool:
        return not await super().__call__(event, state)


matcher = on(
    rule=(
        Rule(NotStartswithRule(tuple(unmatch_beginning)))
        & check_messages
        & (check_to_me if plugin_config.dcqq_relay_only_to_me else None)
    ),
    priority=2,
)


@driver.on_bot_connect
async def get_dc_bot(bot: dc_Bot):
    global dc_bot
    dc_bot = bot


@driver.on_bot_connect
async def get_qq_bot(bot: qq_Bot):
    global qq_bot
    qq_bot = bot


@driver.on_bot_connect
async def get_webhooks(bot: dc_Bot):
    logger.info("prepare webhooks: start")
    global with_webhook_links
    task = [get_webhook(bot, link) for link in without_webhook_links]
    links = await asyncio.gather(*task)
    with_webhook_links.extend(
        link for link in links if isinstance(link, LinkWithWebhook)
    )
    logger.info("prepare webhooks: done")
    if failed := [link for link in links if isinstance(link, int)]:
        logger.error(
            f"{len(failed)} channels failed to get or create webhook: {failed}"
        )


@matcher.handle()
async def message_relay(
    bot: Union[qq_Bot, dc_Bot],
    event: Union[
        GroupMessageEvent,
        GuildMessageCreateEvent,
        GroupRecallNoticeEvent,
        GuildMessageDeleteEvent,
    ],
):
    logger.debug("message relay: start")
    for try_times in range(3):
        try:
            if isinstance(bot, qq_Bot) and isinstance(event, GroupMessageEvent):
                await create_qq_to_dc(bot, event, dc_bot, with_webhook_links)
            elif isinstance(bot, dc_Bot) and isinstance(event, GuildMessageCreateEvent):
                await create_dc_to_qq(bot, event, qq_bot, with_webhook_links)
            elif isinstance(bot, qq_Bot) and isinstance(event, GroupRecallNoticeEvent):
                await delete_qq_to_dc(event, dc_bot, with_webhook_links, just_delete)
            elif isinstance(bot, dc_Bot) and isinstance(event, GuildMessageDeleteEvent):
                await delete_dc_to_qq(event, qq_bot, just_delete)
            else:
                logger.error(
                    "bot type and event type not match: "
                    + f"bot - {bot.type}, event - {type(event)}"
                )
            logger.debug("message relay: done")
            break
        except NameError as e:
            logger.warning(f"message relay error: {e}, retry {try_times + 1}")
            if try_times == 3:
                raise e
            await asyncio.sleep(5)
    else:
        logger.error("message relay: failed")



from typing import Optional

from nonebot import get_plugin_config
from pydantic import BaseModel


class Link(BaseModel):
    dc_guild_id: int
    dc_channel_id: int
    qq_group_id: int


class LinkWithoutWebhook(Link):
    webhook_id: Optional[int] = None
    webhook_token: Optional[str] = None


class LinkWithWebhook(Link):
    webhook_id: int
    webhook_token: str


class Config(BaseModel):
    dcqq_relay_channel_links: list[LinkWithoutWebhook] = []
    """QQ群绑定"""
    dcqq_relay_unmatch_beginning: list[str] = ["/"]
    """不转发的消息开头"""
    dcqq_relay_only_to_me: bool = False
    discord_proxy: Optional[str] = None


plugin_config = get_plugin_config(Config)



from nonebot_plugin_orm import Model
from sqlalchemy import BigInteger
from sqlalchemy.orm import Mapped, mapped_column


class MsgID(Model):
    id: Mapped[int] = mapped_column(primary_key=True)
    dcid: Mapped[int] = mapped_column(type_=BigInteger())
    qqid: Mapped[int]
